@@ -16,7 +16,8 @@ use super::{
     prelude::*,
     scheduled_execution::evaluate_schedule_adjust_numerical_scale,
     scheduled_execution::{
-        circuit_to_schedule, circuit_to_schedule_naive_toposort, evaluate_schedule, SchedulingError,
+        circuit_to_schedule, circuit_to_schedule_naive_toposort, evaluate_schedule, Schedule,
+        SchedulingError,
     },
     CircuitNode, CircuitRc, HashBytes,
 };
@@ -58,6 +59,13 @@ pub struct OptimizationSettings {
     pub create_tests: bool,
 }
 
+#[cfg(not(feature = "real-pyo3"))]
+impl<'source> FromPyObject<'source> for OptimizationSettings {
+    fn extract(_: &'source PyAny) -> PyResult<Self> {
+        unimplemented!()
+    }
+}
+
 impl Default for OptimizationSettings {
     fn default() -> Self {
         Self {
@@ -121,7 +129,7 @@ impl OptimizationSettings {
     }
 }
 
-#[derive(Debug, Clone)]
+#[derive(Debug, Clone, Default)]
 pub struct OptimizationCache {
     pub simplified: HashMap<HashBytes, CircuitRc>,
     pub distributed: HashMap<HashBytes, Option<CircuitRc>>,
@@ -129,16 +137,6 @@ pub struct OptimizationCache {
     pub sum_of_node_sizes: HashMap<HashBytes, BigUint>,
 }
 
-impl Default for OptimizationCache {
-    fn default() -> Self {
-        Self {
-            simplified: HashMap::new(),
-            distributed: HashMap::new(),
-            flops: HashMap::new(),
-            sum_of_node_sizes: HashMap::new(),
-        }
-    }
-}
 #[derive(Debug, Clone, Default)]
 pub struct OptimizationContext {
     pub cache: OptimizationCache,
@@ -212,10 +210,10 @@ pub fn optimize_circuit(circuit: CircuitRc, context: &mut OptimizationContext) -
 }
 
 #[pyfunction]
-pub fn optimize_and_evaluate(
+pub fn optimize_to_schedule(
     circuit: CircuitRc,
     settings: OptimizationSettings,
-) -> Result<Tensor, SchedulingError> {
+) -> Result<Schedule, SchedulingError> {
     let mut context = OptimizationContext {
         settings,
         cache: Default::default(),
@@ -241,18 +239,17 @@ pub fn optimize_and_evaluate(
     if verbose > 1 {
         println!("{}", schedule.get_stats());
     }
-    let eval = || {
-        if settings.adjust_numerical_scale {
-            evaluate_schedule_adjust_numerical_scale(schedule, settings)
-                [&optimized_circuit.info().hash]
-                .clone()
-        } else {
-            evaluate_schedule(schedule)[&optimized_circuit.info().hash].clone()
-        }
-    };
-    let out = timed!(eval(), 10, verbose >= 1);
+    Ok(schedule)
+}
 
-    Ok(out)
+#[pyfunction]
+pub fn optimize_and_evaluate(
+    circuit: CircuitRc,
+    settings: OptimizationSettings,
+) -> Result<Tensor, SchedulingError> {
+    let schedule = optimize_to_schedule(circuit, settings)?;
+
+    Ok(schedule.evaluate(settings))
 }
 
 /// in python, lots of functions take in collections of circuits and operate on them at once
@@ -266,25 +263,22 @@ pub fn optimize_and_evaluate_many(
     circuits: Vec<CircuitRc>,
     settings: OptimizationSettings,
 ) -> Result<Vec<Tensor>, SchedulingError> {
+    let schedule = optimize_to_schedule_many(circuits, settings)?;
+    Ok(schedule.evaluate_many(settings))
+}
+
+#[pyfunction]
+pub fn optimize_to_schedule_many(
+    circuits: Vec<CircuitRc>,
+    settings: OptimizationSettings,
+) -> Result<Schedule, SchedulingError> {
     let flat_concatted = flat_concat(circuits.clone()).rc();
-    let evaluated = optimize_and_evaluate(flat_concatted, settings)?;
-    let out = Python::with_gil(|py| {
-        UN_FLAT_CONCAT
-            .call(
-                py,
-                (
-                    evaluated,
-                    circuits
-                        .iter()
-                        .map(|x| x.info().shape.clone())
-                        .collect::<Vec<Vec<usize>>>(),
-                ),
-                None,
-            )
-            .unwrap()
-            .extract(py)
-            .unwrap()
-    });
-
-    Ok(out)
+    let mut schedule = optimize_to_schedule(flat_concatted, settings)?;
+    schedule.split_shapes = Some(
+        circuits
+            .iter()
+            .map(|x| x.info().shape.clone())
+            .collect::<Vec<Vec<usize>>>(),
+    );
+    Ok(schedule)
 }
