@@ -758,6 +758,29 @@ impl CircuitNode for Rearrange {
                 .map(|x| x.extract(py).unwrap())
         })
     }
+
+    fn intermediate_cost_bound(&self) -> usize {
+        // Most rearranges are free, but some involve a copy
+        //
+        // Cases include:
+        // 1. Flattening axes out of order: a b -> (b a) is a copy unless either dimension is a singleton
+        // 2. Flattening a repeat: a -> (a 2) is a copy
+        // 3. Flattening a view with an offset/repeat:
+        //    * If x = ones(4,4) and y = x[:, 1:] and z = repeat(y, 'a b -> (a b)'), z is a copy
+        //    * If x = ones(4) and y = repeat(x, 'a -> 2 a') and z = repeat(y, 'a b -> (a b)'), z is a copy
+        //
+        // Because of the third case, we cannot confirm that a rearrange returns a view
+        // and has cost 0 unless we have the stride and offset of the child nodes
+        //
+        // Thus, for now, we always return the upper bound cost of numel. We'll explicitly
+        // track and propagate strides alongside shapes later to optimize this
+        if self.spec.output_ints.iter().any(|x| x.len() > 1) {
+            self.info().numel_usize()
+        } else {
+            // some flattens / repeats dont need copy, but we'd need stride to know that
+            0
+        }
+    }
 }
 
 impl CircuitNodeAutoName for Rearrange {
@@ -822,6 +845,25 @@ impl Index {
             ]),
             _ => {}
         }
+        for (idx, l) in zip(&index.0, &node.info().shape) {
+            match idx {
+                TensorAxisIndex::Single(i) => {
+                    if *i < -(*l as i64) || i >= &(*l as i64) {
+                        return Err(CircuitConstructionError::IndexOutOfBounds {
+                            index: index.clone(),
+                            shape: node.info().shape.clone(),
+                        });
+                    }
+                }
+                TensorAxisIndex::Slice(slice) => {
+                    // panic on invalid
+                    slice.stop_u(*l);
+                    slice.start_u(*l);
+                }
+
+                _ => {}
+            }
+        }
         let index = TensorIndex(
             index
                 .0
@@ -909,6 +951,22 @@ impl CircuitNode for Index {
         })
     }
 
+    fn intermediate_cost_bound(&self) -> usize {
+        // Index returns a view if all of the axis indices are not int tensors
+        let has_tensor = self
+            .index
+            .0
+            .iter()
+            .any(|x| matches!(x, TensorAxisIndex::Tensor(_)));
+
+        if has_tensor {
+            // It might still return a view with int tensors in some cases, but we don't handle this
+            self.info().numel_usize()
+        } else {
+            0
+        }
+    }
+
     fn device_dtype_extra<'a>(&'a self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + 'a> {
         Box::new(self.index.0.iter().filter_map(|x| match x {
             TensorAxisIndex::Tensor(tensor) if !tensor.shape().is_empty() => {
@@ -965,7 +1023,9 @@ impl Concat {
         for node in nodes.iter() {
             for (i, s) in node.info().shape.iter().enumerate() {
                 if i != axis && *s != node_shape[i] {
-                    return Err(CircuitConstructionError::ConcatShapeDifferent {});
+                    return Err(CircuitConstructionError::ConcatShapeDifferent {
+                        shapes: nodes.iter().map(|x| x.info().shape.clone()).collect(),
+                    });
                 }
             }
         }
@@ -1227,6 +1287,10 @@ impl CircuitNode for Scatter {
                 .map(|x| x.extract(py).unwrap())
         })
     }
+
+    fn intermediate_cost_bound(&self) -> usize {
+        self.info().numel_usize()
+    }
 }
 
 impl CircuitNodeAutoName for Scatter {
