@@ -79,6 +79,13 @@ impl CircuitNode for ArrayConstant {
         Ok(self.value.clone())
     }
 
+    fn intermediate_cost_bound(&self) -> usize {
+        // Array constants are already allocated, and the compiler's caller keeps a reference to them
+        // in Python, so we can't deallocate them. Thus there's no scheduling optimization we can do
+        // for them.
+        0
+    }
+
     fn device_dtype_extra<'a>(&'a self) -> Box<dyn Iterator<Item = TorchDeviceDtypeOp> + 'a> {
         Box::new(std::iter::once(
             TorchDeviceDtype::from_tensor(&self.value).into(),
@@ -207,6 +214,10 @@ impl CircuitNode for Symbol {
             circuit: self.clone().rc(),
         })
     }
+
+    fn intermediate_cost_bound(&self) -> usize {
+        0
+    }
 }
 
 impl Symbol {
@@ -285,6 +296,11 @@ impl CircuitNode for ScalarConstant {
                 .unwrap())
         })
     }
+
+    fn intermediate_cost_bound(&self) -> usize {
+        // scalar_to_tensor builds a 1 element tensor with a stride 0 view
+        1
+    }
 }
 
 impl ScalarConstant {
