@@ -1,13 +1,9 @@
 use std::collections::{HashMap, HashSet};
 
 use crate::pyo3_prelude::*;
-use itertools::Itertools;
 use num_bigint::BigUint;
 
-use crate::{
-    circuit::diag_rewrite::{add_pull_diags, einsum_push_down_trace},
-    util::{apply_fn_until_same, mapping_until_end, AsOp},
-};
+use crate::util::{apply_fn_until_same, mapping_until_end, AsOp};
 
 use super::{
     algebraic_rewrite::*,
@@ -19,11 +15,13 @@ use super::{
         index_concat_drop_unreached,
     },
     deep_map_unwrap, deep_map_unwrap_preorder,
+    diag_rewrite::{add_pull_diags, einsum_push_down_trace},
+    prelude::*,
     scatter_rewrite::{
         add_pull_scatter, einsum_pull_scatter, index_einsum_to_scatter, scatter_elim_identity,
         scatter_fuse, scatter_pull_removable_axes,
     },
-    visit_circuit, Add, Circuit, CircuitNode, CircuitRc, Concat, GeneralFunction, HashBytes, Index,
+    visit_circuit, Add, Concat, GeneralFunction, HashBytes, Index,
 };
 
 /// seperate _py function because pyfunctions cant take reference arguments
@@ -229,12 +227,12 @@ pub fn deep_push_down_index(circ: CircuitRc, min_size: Option<usize>) -> Circuit
 pub fn deep_heuristic_nest_adds(circ: CircuitRc) -> CircuitRc {
     let circ = deep_canonicalize(circ);
     let mut seen_adds: HashSet<Add> = HashSet::new();
-    let mut mapping: HashMap<Add, Add> = HashMap::new();
     visit_circuit(&circ, &mut |c: &Circuit| {
         if let Some(add) = c.as_add() {
             seen_adds.insert(add.clone());
         }
     });
+    // TODO: profile and optimize these
     let mut intersections: HashSet<Add> = HashSet::new();
     for circ in &seen_adds {
         for circ2 in &seen_adds {
@@ -247,20 +245,21 @@ pub fn deep_heuristic_nest_adds(circ: CircuitRc) -> CircuitRc {
                 None,
             )
             .unwrap();
-            if intersections.len() >= 2 && &intersection != circ && &intersection != circ2 {
+            if intersection.nodes.len() >= 2 && &intersection != circ && &intersection != circ2 {
                 intersections.insert(intersection);
             }
         }
     }
     seen_adds.extend(intersections);
-    while let Some((sub, sup)) = (|| {
+    let mut mapping: HashMap<Add, Add> = HashMap::new();
+    while let Some((sup, new_sup)) = (|| {
         for cand_sub in &seen_adds {
-            if cand_sub.nodes.iter().unique().count() > 1 {
+            if cand_sub.nodes.len() >= 2 {
                 for cand_sup in &seen_adds {
-                    if cand_sub != cand_sup
-                        && cand_sub.nodes.iter().all(|x| cand_sup.nodes.contains(x))
-                    {
-                        return Some((cand_sub.clone(), cand_sup.clone()));
+                    if cand_sub != cand_sup {
+                        if let Some(new) = extract_add(cand_sup, cand_sub) {
+                            return Some((cand_sup.clone(), new.clone()));
+                        }
                     }
                 }
             }
@@ -268,25 +267,12 @@ pub fn deep_heuristic_nest_adds(circ: CircuitRc) -> CircuitRc {
         None
     })() {
         seen_adds.remove(&sup);
-        let new_sup = canonicalize_node(
-            &Add::try_new(
-                sup.nodes
-                    .iter()
-                    .cloned()
-                    .filter(|x| !sub.nodes.contains(x))
-                    .chain(std::iter::once(sub.clone().rc()))
-                    .collect(),
-                None,
-            )
-            .unwrap()
-            .rc(),
-        );
-        let new_sup = new_sup.as_add().unwrap();
         seen_adds.insert(new_sup.clone());
         mapping.insert(sup, new_sup.clone());
     }
+
     deep_map_unwrap_preorder(&circ, |c| {
-        c.map_or_clone(|add| {
+        c.map_or_clone(|add: &Add| {
             let add = mapping_until_end(add, &mapping);
 
             if add.info().numel() > BigUint::from(100_000_000usize) {
@@ -299,21 +285,17 @@ pub fn deep_heuristic_nest_adds(circ: CircuitRc) -> CircuitRc {
 }
 
 pub fn add_nest_ltr(add: &Add) -> Add {
-    let mut result = add.clone();
-    while result.nodes.len() > 2 {
-        result = Add::try_new(
-            std::iter::once(
-                Add::try_new(vec![result.nodes[0].clone(), result.nodes[1].clone()], None)
-                    .unwrap()
-                    .rc(),
-            )
-            .chain(result.nodes.iter().dropping(2).cloned())
-            .collect::<Vec<CircuitRc>>(),
-            None,
-        )
-        .unwrap();
-    }
-    result
+    let (l, r) = add.nodes.split_at(2.min(add.nodes.len()));
+    let base = Add::try_new(l.to_vec(), None).unwrap();
+    r.into_iter().fold(base, |acc, x| {
+        Add::try_new(vec![acc.rc(), x.clone()], None).unwrap()
+    })
+}
+
+#[pyfunction]
+#[pyo3(name = "add_nest_ltr")]
+pub fn add_nest_ltr_py(add: Add) -> Add {
+    add_nest_ltr(&add)
 }
 
 #[pyfunction]
