@@ -2,11 +2,7 @@ use crate::all_imports::TorchDeviceDtypeOp;
 use crate::pyo3_prelude::*;
 use std::{collections::HashMap, iter::zip};
 
-use super::circuit_optimizer::optimize_and_evaluate;
-use super::{
-    Add, ArrayConstant, CachedCircuitInfo, Circuit, CircuitConstructionError, CircuitNode,
-    CircuitNodeAutoName, CircuitRc, Einsum, HashBytes,
-};
+use super::{prelude::*, CachedCircuitInfo, HashBytes};
 use crate::py_types::{GENERALFUNCTIONS, PY_NONE};
 use crate::{
     circuit::PyCircuitBase,
@@ -353,6 +349,9 @@ impl GeneralFunction {
 #[test]
 #[ignore] // needs torch which ci rust test doesn't have...should add that
 fn test_generalfuction() {
+    use super::{Add, ArrayConstant, Einsum};
+    use crate::circuit::circuit_optimizer::optimize_and_evaluate;
+
     pyo3::prepare_freethreaded_python();
     let circuit = Einsum::nrc(
         vec![(
