@@ -1,5 +1,6 @@
 #![allow(clippy::borrow_deref_ref)]
-use crate::pyo3_prelude::*;
+use crate::py_types::PY_MAKE_BYTES;
+use crate::{pycall, pyo3_prelude::*};
 use num_bigint::BigUint;
 use pyo3::{exceptions, pyclass::CompareOp};
 use std::hash::Hash;
@@ -10,6 +11,7 @@ use std::{
     rc::Rc,
 };
 
+use macro_rules_attribute::apply;
 use std::fmt::Debug;
 use thiserror::Error;
 
@@ -37,7 +39,7 @@ pub mod scheduled_execution;
 pub mod scheduling_z3;
 
 use crate::{
-    cached_lambda, caching_struct,
+    cached_lambda,
     py_types::{PyShape, Tensor, CIRCUIT_RUST_TO_PY},
     rearrange_spec::{OpSize, RearrangeSpec},
     tensor_util::{Shape, TensorAxisIndex, TensorIndex, TorchDeviceDtype, TorchDeviceDtypeOp},
@@ -115,6 +117,12 @@ pub trait CircuitNode: CircuitNodePrivate {
         device_dtype: &TorchDeviceDtype,
     ) -> Result<Tensor, TensorEvalError>;
 
+    /// At most how many elements will evaluating this circuit require allocating
+    /// new memory (that we are allowed to free ourselves) for? Used to improve scheduling.
+    fn intermediate_cost_bound(&self) -> usize {
+        self.info().numel_usize()
+    }
+
     // ==== default section ===
     // FUNCTIONS BELOW HERE *shouldn't* be overridden by implementors!
     // (if you do implement, this won't be picked up on by union types!)
@@ -364,7 +372,7 @@ macro_rules! define_circuit_union_impl {
 
         #[cfg(not(feature = "real-pyo3"))]
         impl<'source> $crate::pyo3::FromPyObject<'source> for $name {
-            fn extract(circuit_obj: &'source $crate::pyo3::PyAny) -> $crate::pyo3::PyResult<Self> {
+            fn extract(_: &'source $crate::pyo3::PyAny) -> $crate::pyo3::PyResult<Self> {
                 unimplemented!()
             }
         }
@@ -683,9 +691,21 @@ impl<T: CircuitNodeUnion> CircuitNode for T {
     ) -> Result<Tensor, TensorEvalError> {
         self.as_trait_obj().eval_tensors(tensors, device_dtype)
     }
+
+    fn intermediate_cost_bound(&self) -> usize {
+        self.as_trait_obj().intermediate_cost_bound()
+    }
 }
 
 pub type HashBytes = [u8; 32];
+#[derive(Clone, Copy, Hash, PartialEq, Eq)]
+pub struct PyHashBytes(pub [u8; 32]);
+impl IntoPy<PyObject> for PyHashBytes {
+    fn into_py(self, _py: Python<'_>) -> PyObject {
+        pycall!(PY_MAKE_BYTES, (self.0,))
+    }
+}
+
 #[derive(Clone, Default)]
 pub struct CachedCircuitInfo {
     pub shape: Shape,
@@ -709,6 +729,15 @@ impl CachedCircuitInfo {
     pub fn numel(&self) -> BigUint {
         self.shape.iter().map(|x| BigUint::from(*x)).product()
     }
+    /// Saturating element count
+    pub fn numel_usize(&self) -> usize {
+        let numel_digits = self.numel().to_u64_digits();
+        if numel_digits.len() == 1 {
+            numel_digits[0] as usize
+        } else {
+            usize::MAX
+        }
+    }
     pub fn rank(&self) -> usize {
         self.shape.len()
     }
@@ -757,8 +786,8 @@ pub enum CircuitConstructionError {
     #[error("Concat requires at least one node")]
     ConcatZeroNodes {},
 
-    #[error("Concat nodes have different shapes")]
-    ConcatShapeDifferent {},
+    #[error("Concat nodes have different shapes {shapes:?}")]
+    ConcatShapeDifferent { shapes: Vec<Shape> },
 
     #[error("index rank too high: {index_rank} vs {node_rank}")]
     IndexRankTooHigh { index_rank: usize, node_rank: usize },
@@ -840,64 +869,18 @@ pub fn make_children_zero<T: CircuitNode>(circuit: &T) -> T {
     })
 }
 
-caching_struct!(
-    // Maybe this should actually own function? idk
-    pub struct DeepMapUnwrapC<{'a}, {F : Fn(&Circuit) -> CircuitRc}> { f : &'a F }
-    [pub cache : T] pub fn map(&mut self_, circuit: Circuit) -> CircuitRc = {
-        let inner_mapped =
-            circuit.map_children_unwrap(&mut |inner: &Circuit| self_.map(inner.clone()));
-        (self_.f)(&inner_mapped)
-    }
-);
-
-/// Like [cached_lambda](crate::cached_lambda), but assumes all inputs are
-/// circuit types and uses the bytes as the hash key. If you need more
-/// flexiblity, specify a custom Key using [cached_lambda](crate::cached_lambda).
-///
-/// TODO: maybe we should support generic lambda ish stuff somehow? (seems a bit cursed imo...)
-///
-/// # Example
-///
-/// ```
-/// # use rust_circuit::{cached_circuit_lambda, circuit::{ScalarConstant, Circuit, CircuitNode, HashBytes}};
-/// let starting_at = 3;
-///
-/// cached_circuit_lambda! {
-///     let mut func = |x: &Circuit| -> String = {
-///         x.name().unwrap()[starting_at..].to_owned()
-///     };
-/// };
-///
-/// let circ: Circuit = ScalarConstant::new(0.7, vec![8], Some("name".to_owned())).into();
-/// assert_eq!(func(&circ), "e".to_owned());
-/// ```
-#[macro_export]
-macro_rules! cached_circuit_lambda {
-    {
-        let mut $func_id:ident = |$($circuit_arg:ident : $circuit_type : ty),* $(,)?| -> $ret:ty  = $body:expr;
-    } => {
-        paste::paste! {
-            type [<$func_id:camel MacroKeyType>] = $crate::repeated_tuple_impl!({HashBytes} ($($circuit_arg,)*));
-            $crate::cached_lambda!{
-                {Key : [<$func_id:camel MacroKeyType>] = ($($circuit_arg.info().hash,)*)}
-                let mut $func_id = |$($circuit_arg : $circuit_type,)*| -> $ret = $body;
-            };
-        }
-    }
-}
-
 pub fn evaluate_fn(circ: &Circuit) -> Result<Tensor, TensorEvalError> {
     let device_dtype = get_compatible_dtype(circ);
-    cached_lambda! {
-        {Key : HashBytes = circ.info().hash}
-        let mut recurse = |circ: Circuit| -> Result<Tensor, TensorEvalError> = {
-            let child_tensors: Result<Vec<Tensor>, TensorEvalError> =
-                circ.children().map(|x| recurse((**x).clone())).collect();
-            let child_tensors = child_tensors?;
+    #[apply(cached_lambda)]
+    #[key(circ.info().hash, HashBytes)]
+    fn recurse(circ: Circuit) -> Result<Tensor, TensorEvalError> {
+        let child_tensors: Result<Vec<Tensor>, TensorEvalError> =
+            circ.children().map(|x| recurse((**x).clone())).collect();
+        let child_tensors = child_tensors?;
 
-            circ.eval_tensors(&child_tensors, &device_dtype)
-        };
+        circ.eval_tensors(&child_tensors, &device_dtype)
     }
+
     recurse(circ.clone())
 }
 
@@ -905,12 +888,11 @@ pub fn deep_map<F>(circuit: &Circuit, f: F) -> Result<CircuitRc, CircuitConstruc
 where
     F: Fn(&Circuit) -> Result<CircuitRc, CircuitConstructionError>,
 {
-    cached_circuit_lambda! {
-        let mut recurse = |circ: &Circuit| -> Result<CircuitRc,CircuitConstructionError> = {
-            let inner_mapped =
-                circ.map_children(&mut recurse)?;
-            f(&inner_mapped)
-        };
+    #[apply(cached_lambda)]
+    #[key(circ.info().hash, HashBytes)]
+    fn recurse(circ: &Circuit) -> Result<CircuitRc, CircuitConstructionError> {
+        let inner_mapped = circ.map_children(&mut recurse)?;
+        f(&inner_mapped)
     }
     recurse(circuit)
 }
@@ -919,10 +901,10 @@ pub fn deep_map_preorder<F>(circuit: &Circuit, f: F) -> Result<CircuitRc, Circui
 where
     F: Fn(&Circuit) -> Result<CircuitRc, CircuitConstructionError>,
 {
-    cached_circuit_lambda! {
-        let mut recurse = |circ: &Circuit| -> Result<CircuitRc,CircuitConstructionError> = {
-            f(circ)?.map_children(&mut recurse).map(|z|z.rc())
-        };
+    #[apply(cached_lambda)]
+    #[key(circ.info().hash, HashBytes)]
+    fn recurse(circ: &Circuit) -> Result<CircuitRc, CircuitConstructionError> {
+        f(circ)?.map_children(&mut recurse).map(|z| z.rc())
     }
     recurse(circuit)
 }
@@ -931,12 +913,11 @@ pub fn deep_map_unwrap<F>(circuit: &Circuit, f: F) -> CircuitRc
 where
     F: Fn(&Circuit) -> CircuitRc,
 {
-    cached_circuit_lambda! {
-        let mut recurse = |circ: &Circuit| -> CircuitRc = {
-            let inner_mapped =
-                circ.map_children_unwrap(&mut recurse);
-            f(&inner_mapped)
-        };
+    #[apply(cached_lambda)]
+    #[key(circ.info().hash, HashBytes)]
+    fn recurse(circ: &Circuit) -> CircuitRc {
+        let inner_mapped = circ.map_children_unwrap(&mut recurse);
+        f(&inner_mapped)
     }
     recurse(circuit)
 }
@@ -946,15 +927,15 @@ pub fn deep_map_unwrap_preorder<F>(circuit: &Circuit, f: F) -> CircuitRc
 where
     F: Fn(&Circuit) -> CircuitRc,
 {
-    cached_circuit_lambda! {
-        let mut recurse = |circ: &Circuit| -> CircuitRc = {
-            f(circ).map_children_unwrap(&mut recurse).rc()
-        };
+    #[apply(cached_lambda)]
+    #[key(circ.info().hash, HashBytes)]
+    fn recurse(circ: &Circuit) -> CircuitRc {
+        f(circ).map_children_unwrap(&mut recurse).rc()
     }
     recurse(circuit)
 }
 
-/// visits each subcircuit that's in any circuit once!
+/// visits each subcircuit that's in any circuit once! (pre-order traversal if it's a tree)
 pub fn visit_circuit<F>(circuit: &Circuit, mut f: F)
 where
     F: FnMut(&Circuit),
@@ -1023,13 +1004,13 @@ pub fn deep_map_op<F>(circuit: &Circuit, f: F) -> Option<CircuitRc>
 where
     F: Fn(&Circuit) -> Option<CircuitRc>,
 {
-    cached_circuit_lambda! {
-        let mut recurse = |circ: &Circuit| -> Option<CircuitRc> = {
-            let inner_mapped = circ.map_children_op(&mut recurse);
-            inner_mapped
-                .map(|x| x.and_then_or_clone(&f))
-                .or_else(|| f(circ))
-        };
+    #[apply(cached_lambda)]
+    #[key(circ.info().hash, HashBytes)]
+    fn recurse(circ: &Circuit) -> Option<CircuitRc> {
+        let inner_mapped = circ.map_children_op(&mut recurse);
+        inner_mapped
+            .map(|x| x.and_then_or_clone(&f))
+            .or_else(|| f(circ))
     }
     recurse(circuit)
 }
@@ -1038,12 +1019,12 @@ pub fn deep_map_pre_new_children<F>(circuit: &Circuit, f: F) -> CircuitRc
 where
     F: Fn(&Circuit, &Vec<CircuitRc>) -> CircuitRc,
 {
-    cached_circuit_lambda! {
-        let mut recurse = |circ: &Circuit| -> CircuitRc = {
-            let old_children:Vec<CircuitRc>=circ.children().collect();
-            let new_children = old_children.iter().map(|x|&***x).map(recurse).collect();
-            f(&circ,&new_children)
-        };
+    #[apply(cached_lambda)]
+    #[key(circ.info().hash, HashBytes)]
+    fn recurse(circ: &Circuit) -> CircuitRc {
+        let old_children: Vec<CircuitRc> = circ.children().collect();
+        let new_children = old_children.iter().map(|x| &***x).map(recurse).collect();
+        f(&circ, &new_children)
     }
     recurse(circuit)
 }
@@ -1052,17 +1033,23 @@ pub fn deep_map_op_pre_new_children<F>(circuit: &Circuit, f: F) -> Option<Circui
 where
     F: Fn(&Circuit, &Vec<CircuitRc>) -> Option<CircuitRc>,
 {
-    cached_circuit_lambda! {
-        let mut recurse = |circ: &Circuit| -> Option<CircuitRc> = {
-            let old_children:Vec<CircuitRc>=circ.children().collect();
-            let new_children:Vec<Option<CircuitRc>> = old_children.iter().map(|x|&***x).map(recurse).collect();
-            if new_children.iter().all(|x|x.is_none()){
-                f(circ,&old_children)
-            }else{
-                let new_real_children = zip(old_children,new_children).map(|(old,new)|new.unwrap_or(old)).collect();
-                Some(f(circ,&new_real_children).unwrap_or_else(||circ.map_children_unwrap_idxs(|i|new_real_children[i].clone()).rc()))
-            }
-        };
+    #[apply(cached_lambda)]
+    #[key(circ.info().hash, HashBytes)]
+    fn recurse(circ: &Circuit) -> Option<CircuitRc> {
+        let old_children: Vec<CircuitRc> = circ.children().collect();
+        let new_children: Vec<Option<CircuitRc>> =
+            old_children.iter().map(|x| &***x).map(recurse).collect();
+        if new_children.iter().all(|x| x.is_none()) {
+            f(circ, &old_children)
+        } else {
+            let new_real_children = zip(old_children, new_children)
+                .map(|(old, new)| new.unwrap_or(old))
+                .collect();
+            Some(f(circ, &new_real_children).unwrap_or_else(|| {
+                circ.map_children_unwrap_idxs(|i| new_real_children[i].clone())
+                    .rc()
+            }))
+        }
     }
     recurse(circuit)
 }
@@ -1242,6 +1229,11 @@ impl PyCircuitBase {
         self.0.name().unwrap_or("")
     }
 
+    #[getter]
+    fn intermediate_cost_bound(&self) -> usize {
+        self.0.intermediate_cost_bound()
+    }
+
     // TODO: probably could be more efficient...
     fn children(&self) -> Vec<CircuitRc> {
         self.0.children().collect()
@@ -1252,8 +1244,8 @@ impl PyCircuitBase {
     }
 
     #[getter]
-    fn hash(&self) -> [u8; 32] {
-        self.info().hash
+    fn hash(&self) -> PyHashBytes {
+        PyHashBytes(self.info().hash)
     }
 
     fn __hash__(&self) -> u64 {
