@@ -5,9 +5,11 @@ use std::{
 
 use crate::{
     cached_lambda,
-    py_types::{ExtraPySelfOps, Tensor, PY_SCALAR_TO_TENSOR, TENSOR_SCALE},
+    circuit::circuit_optimizer::OptimizationContext,
+    py_types::{ExtraPySelfOps, Tensor, PY_SCALAR_TO_TENSOR, TENSOR_SCALE, UN_FLAT_CONCAT},
     pycall,
     tensor_util::{Shape, TorchDeviceDtype},
+    timed,
 };
 
 use super::{
@@ -17,10 +19,12 @@ use super::{
     prelude::*,
     print::oom_fmt,
     scheduling_z3::schedule_dag_strategy_ints,
-    visit_circuit, ArrayConstant, HashBytes, ScalarConstant, Symbol,
+    visit_circuit, ArrayConstant, HashBytes, PyHashBytes, ScalarConstant, Symbol,
 };
 use crate::pyo3_prelude::*;
 use crate::util::filter_to_idx;
+use itertools::Itertools;
+use macro_rules_attribute::apply;
 use num_bigint::BigUint;
 use pyo3::{create_exception, exceptions::PyException};
 use thiserror::Error;
@@ -31,23 +35,87 @@ pub enum Instruction {
     Compute(CircuitRc),
 }
 
-// impl IntoPy<PyObject> for Instruction {
-//     fn into_py(self, py: Python<'_>) -> PyObject {
-//         ().into_py(py)
-//     }
-// }
+impl IntoPy<PyObject> for Instruction {
+    fn into_py(self, py: Python<'_>) -> PyObject {
+        match self {
+            Instruction::Drop(hb) => PyHashBytes(hb).into_py(py),
+            Instruction::Compute(c) => c.into_py(py),
+        }
+    }
+}
 
+#[pyclass(unsendable)]
 #[derive(Clone, Debug)]
 pub struct Schedule {
-    instructions: Vec<Instruction>,
-    constants: HashMap<HashBytes, Tensor>,
+    #[pyo3(get)]
+    pub instructions: Vec<Instruction>,
+    pub constants: HashMap<HashBytes, Tensor>,
     // keep scalar constants seperate so adjust numerical scale can work without losing precision
     // before when these were in tensors, they had to be right dtype and therefore overflowed when
     // adjustment needed
-    scalar_constants: HashMap<HashBytes, ScalarConstant>,
-    device_dtype: TorchDeviceDtype,
+    pub scalar_constants: HashMap<HashBytes, ScalarConstant>,
+    #[pyo3(get)]
+    pub device_dtype: TorchDeviceDtype,
+    pub output_circuit: Option<CircuitRc>,
+    pub split_shapes: Option<Vec<Shape>>,
 }
+#[pymethods]
 impl Schedule {
+    pub fn replace_tensors(&self, map: HashMap<HashBytes, Tensor>) -> PyResult<Self> {
+        let mut result = self.clone();
+        for (k, v) in map {
+            if !result.constants.contains_key(&k) {
+                return Err(PyErr::new::<pyo3::exceptions::PyKeyError, _>(
+                    "key circuit wasn't present in original",
+                ));
+            }
+            result.constants.insert(k, v);
+        }
+        Ok(result)
+    }
+
+    #[pyo3(name = "map_tensors")]
+    pub fn map_tensors_py(&self, f: PyObject) -> PyResult<Self> {
+        let mut result = self.clone();
+        result.constants = result
+            .constants
+            .into_iter()
+            .map(|(hash, tensor)| (hash, pycall!(f, (PyHashBytes(hash), tensor))))
+            .collect();
+        Ok(result)
+    }
+
+    #[getter]
+    #[pyo3(name = "constants")]
+    pub fn constants_py(&self) -> HashMap<PyHashBytes, Tensor> {
+        self.constants
+            .iter()
+            .map(|(k, v)| (PyHashBytes(*k), v.clone()))
+            .collect()
+    }
+
+    pub fn evaluate(&self, settings: OptimizationSettings) -> Tensor {
+        let eval = || {
+            if settings.adjust_numerical_scale {
+                evaluate_schedule_adjust_numerical_scale(self, settings)
+                    [&self.output_circuit.as_ref().unwrap().info().hash]
+                    .clone()
+            } else {
+                evaluate_schedule(self)[&self.output_circuit.as_ref().unwrap().info().hash].clone()
+            }
+        };
+        timed!(eval(), 10, settings.verbose >= 1)
+    }
+
+    pub fn evaluate_many(&self, settings: OptimizationSettings) -> Vec<Tensor> {
+        let single = self.evaluate(settings);
+        self.split(single)
+    }
+
+    pub fn split(&self, tensor: Tensor) -> Vec<Tensor> {
+        pycall!(UN_FLAT_CONCAT, (tensor, self.split_shapes.clone().unwrap()))
+    }
+
     pub fn get_stats(&self) -> ScheduleStats {
         let mut mem: BigUint = BigUint::from(0usize);
         let mut max_mem: BigUint = BigUint::from(0usize);
@@ -81,10 +149,14 @@ impl Schedule {
     }
 }
 
+#[pyclass(unsendable)]
 #[derive(Clone, Debug)]
 pub struct ScheduleStats {
+    #[pyo3(get)]
     max_mem: BigUint,
+    #[pyo3(get)]
     constant_mem: BigUint, // this has already been allocated so it can't be over 2^64
+    #[pyo3(get)]
     max_circuit_set: HashSet<CircuitRc>,
 }
 
@@ -143,7 +215,7 @@ impl Display for ScheduleStats {
     }
 }
 
-pub fn evaluate_schedule(schedule: Schedule) -> HashMap<HashBytes, Tensor> {
+pub fn evaluate_schedule(schedule: &Schedule) -> HashMap<HashBytes, Tensor> {
     let mut live: HashMap<HashBytes, Tensor> = schedule.constants.clone();
     live.extend(
         schedule
@@ -206,7 +278,7 @@ pub fn evaluate_schedule(schedule: Schedule) -> HashMap<HashBytes, Tensor> {
 }
 
 pub fn evaluate_schedule_adjust_numerical_scale(
-    schedule: Schedule,
+    schedule: &Schedule,
     settings: OptimizationSettings,
 ) -> HashMap<HashBytes, Tensor> {
     // we store (tensor, scale) where scale is a number the tensor's been multiplied by
@@ -370,8 +442,8 @@ pub fn evaluate_schedule_adjust_numerical_scale(
                         .unwrap(),
                         tensors_and_scales[0].1,
                     ),
-                    _ => {
-                        unimplemented!()
+                    Circuit::ScalarConstant(_) | Circuit::ArrayConstant(_) | Circuit::Symbol(_) => {
+                        panic!("constant found as schedule instruction, not supposed to happen")
                     }
                 };
                 assert!(result.0.shape()[..] == circ.info().shape[..]);
@@ -407,6 +479,8 @@ pub fn order_to_schedule(
             .map(|x| (x.info().hash, x.clone()))
             .collect(),
         device_dtype,
+        split_shapes: None,
+        output_circuit: None,
     };
 
     let mut seen_dependencies: HashSet<HashBytes> = HashSet::new();
@@ -434,6 +508,10 @@ pub fn order_to_schedule(
 pub struct Dag {
     pub children: Vec<Vec<usize>>,
     pub parents: Vec<Vec<usize>>,
+    /// the cost is the number of new elements computing the node entails allocating.
+    /// some nodes compute views into existing memory, such as some indexes and all ArrayConstants
+    /// (as they're already allocated), so they have cost 0
+    /// if a node computes a new tensor, the cost is the product of the shape.
     pub node_costs: Vec<usize>,
     pub node_hashes: Vec<HashBytes>,
     pub hash_to_node: HashMap<HashBytes, usize>,
@@ -442,22 +520,89 @@ impl Dag {
     pub fn get_outputs(&self) -> Vec<usize> {
         filter_to_idx(&self.parents.iter().collect(), &mut |y| y.is_empty())
     }
+    /// Create a new dag with a subset of the nodes. Edges to ignored
+    /// nodes are replaced with edges that pass through to nodes we're keeping
+    pub fn sub_dag(&self, nodes_to_keep: &Vec<usize>) -> Dag {
+        // map indices from nodes_to_keep -> arange
+        let old_to_new: HashMap<usize, usize> = nodes_to_keep
+            .iter()
+            .enumerate()
+            .map(|x| (*x.1, x.0))
+            .collect();
+        // fill out everything except the parents and children
+        // assumes the costs stay the same
+        let mut result = Dag {
+            node_hashes: nodes_to_keep.iter().map(|x| self.node_hashes[*x]).collect(),
+            hash_to_node: nodes_to_keep
+                .iter()
+                .enumerate()
+                .map(|(i, x)| (self.node_hashes[*x], i))
+                .collect(),
+            node_costs: nodes_to_keep.iter().map(|x| self.node_costs[*x]).collect(),
+            children: vec![vec![]; nodes_to_keep.len()],
+            parents: vec![vec![]; nodes_to_keep.len()],
+        };
+        // new children = (old children we're keeping) + (new children of old children we're not keeping)
+        #[apply(cached_lambda)]
+        fn new_children(old_i: usize) -> Vec<usize> {
+            let old_children = &self.children[old_i];
+            old_children
+                .iter()
+                .flat_map(|old_child| {
+                    if let Some(n) = old_to_new.get(old_child) {
+                        vec![*n]
+                    } else {
+                        new_children(*old_child)
+                    }
+                })
+                .collect()
+        }
+        // fill out parents and children
+        for (new_i, old_i) in nodes_to_keep.iter().enumerate() {
+            let children = new_children(*old_i);
+            result.children[new_i] = children.clone();
+            for child in children {
+                result.parents[child].push(new_i);
+            }
+        }
+        result
+    }
+
+    /// Find all of the nontrivial chains within the dag which are internally not connected
+    /// to other nodes, and return the top / parent indices
+    /// This corresponds to all the size >=2 subcircuits in a circuit which involve serial computation
+    /// (Technically these are the nontrivial maximal chain induced subgraphs)
+    pub fn chains(&self) -> Vec<usize> {
+        #[apply(cached_lambda)]
+        fn top_of_chain(node: usize) -> usize {
+            if self.parents[node].len() == 1 && self.children[self.parents[node][0]].len() == 1 {
+                top_of_chain(self.parents[node][0])
+            } else {
+                node
+            }
+        }
+        let node_to_top: Vec<usize> = (0..self.node_costs.len())
+            .map(|node| top_of_chain(node))
+            .collect();
+        node_to_top
+            .into_iter()
+            .enumerate()
+            // filter out length 1 chains
+            .filter_map(|(node, top)| if node == top { None } else { Some(top) })
+            // return each parent of a nontrivial chain once
+            .unique()
+            .collect()
+    }
 }
 
 pub fn circuits_to_dag(circuits: &[&Circuit]) -> Dag {
     let mut result: Dag = Default::default();
+    // append the circuit to the dag if it's not already there, and return its index
     let number_node = |c: &Circuit, result: &mut Dag| -> usize {
         if let Some(idx) = result.hash_to_node.get(&c.info().hash) {
             return *idx;
         }
-        result
-            .node_costs
-            .push(if !matches!(c, Circuit::ArrayConstant(_)) {
-                assert!(c.info().numel() < BigUint::from(usize::MAX));
-                c.info().numel().to_u64_digits()[0] as usize
-            } else {
-                0
-            });
+        result.node_costs.push(c.intermediate_cost_bound());
         result.node_hashes.push(c.info().hash);
         result
             .hash_to_node
@@ -468,6 +613,7 @@ pub fn circuits_to_dag(circuits: &[&Circuit]) -> Dag {
     };
     for circuit in circuits {
         visit_circuit(circuit, &mut |c: &Circuit| {
+            // ArrayConstants are never added to the dag to begin with because they're always 0 cost
             if !matches!(c, Circuit::ArrayConstant(_) | Circuit::ScalarConstant(_)) {
                 let my_number = number_node(c, &mut result);
                 let children_to_consider: Vec<CircuitRc> = c
@@ -493,53 +639,59 @@ pub fn circuits_to_dag(circuits: &[&Circuit]) -> Dag {
     result
 }
 
-fn simplify_dag_kept_nodes(dag: &Dag, chunk_size: usize) -> Vec<usize> {
-    (0..dag.node_costs.len())
-        .filter(|i| dag.node_costs[*i] >= chunk_size)
-        .collect()
-}
-
 fn simplify_dag_for_scheduling_and_kept_nodes(dag: &Dag, chunk_size: usize) -> (Dag, Vec<usize>) {
-    let nodes_to_keep: Vec<usize> = simplify_dag_kept_nodes(dag, chunk_size);
-    let old_to_new: HashMap<usize, usize> = nodes_to_keep
-        .iter()
-        .enumerate()
-        .map(|x| (*x.1, x.0))
+    // filter out nodes that are too small to be worth scheduling (such as cost 0 indexes)
+    let nodes_to_keep: Vec<usize> = (0..dag.node_costs.len())
+        .filter(|i| dag.node_costs[*i] >= chunk_size)
         .collect();
-    let mut result = Dag {
-        node_hashes: nodes_to_keep.iter().map(|x| dag.node_hashes[*x]).collect(),
-        hash_to_node: nodes_to_keep
-            .iter()
-            .enumerate()
-            .map(|(i, x)| (dag.node_hashes[*x], i))
-            .collect(),
-        node_costs: nodes_to_keep.iter().map(|x| dag.node_costs[*x]).collect(),
-        children: vec![vec![]; nodes_to_keep.len()],
-        parents: vec![vec![]; nodes_to_keep.len()],
-    };
-
-    cached_lambda! {
-        let mut important_children = |old_i:usize|->Vec<usize> = {
-            let old_children = &dag.children[old_i];
-            old_children.iter().flat_map(|old_child|{
-                if let Some(n)=old_to_new.get(old_child){
-                    vec![*n]
-                }else{
-                    important_children(*old_child)
-                }
-            }).collect()
-        };
-    }
-
-    for (new_i, old_i) in nodes_to_keep.iter().enumerate() {
-        let children = important_children(*old_i);
-        result.children[new_i] = children.clone();
-        for child in children {
-            result.parents[child].push(new_i);
+    let dag = dag.sub_dag(&nodes_to_keep);
+    // now we filter out some parts in chains of serial computation
+    // if we have a chain of nodes `start -> node_1 -> ... -> end`,
+    // we can simplify the dag to only include the node with the smallest cost,
+    // and the highest cost nodes on either side.
+    // we need the smallest cost node as after computing this is the best time to compute
+    // other nodes from other places in parallel
+    // and we need the maximum nodes on either side so the scheduler has the right bound
+    // on intermediate memory consumption along the serial chain
+    let chain_ends = dag.chains();
+    let mut nodes_to_prune: HashSet<usize> = HashSet::new();
+    for chain_end in chain_ends {
+        // iterate backwards down the chain and accumulate the min and max on either side
+        let c = &dag.node_costs;
+        let mut current = (chain_end, c[chain_end]);
+        let mut min_node = current;
+        let mut max_before = current;
+        let mut max_after = current;
+        while dag.children[current.0].len() == 1 {
+            current = {
+                let child = dag.children[current.0][0];
+                (child, c[child])
+            };
+            if current.1 < min_node.1 {
+                max_before = *vec![max_before, min_node, max_after, current]
+                    .iter()
+                    .max_by_key(|(_node, cost)| cost)
+                    .unwrap();
+                min_node = current;
+                max_after = current;
+            } else if current.1 > max_after.1 {
+                max_after = current;
+            }
+        }
+        // and drop all the other nodes
+        let mut current = chain_end;
+        while dag.children[current].len() == 1 {
+            current = dag.children[current][0];
+            if current != min_node.0 && current != max_before.0 && current != max_after.0 {
+                nodes_to_prune.insert(current);
+            }
         }
     }
-
-    (result, nodes_to_keep)
+    let nodes_to_keep = (0..dag.node_costs.len())
+        .filter(|i| !nodes_to_prune.contains(i))
+        .collect();
+    let dag = dag.sub_dag(&nodes_to_keep);
+    (dag, nodes_to_keep)
 }
 
 fn subset_order_to_full_order(
@@ -622,7 +774,7 @@ pub fn circuit_to_schedule(
         .iter()
         .map(|x| to_node[&dag.node_hashes[*x]].clone())
         .collect();
-    let out = order_to_schedule(
+    let mut out = order_to_schedule(
         &circuit_order,
         &to_node
             .iter()
@@ -638,19 +790,22 @@ pub fn circuit_to_schedule(
             .collect(),
         get_compatible_dtype(&circuit),
     );
+    out.output_circuit = Some(circuit);
 
     Ok(out)
 }
 
 pub fn circuit_to_schedule_naive_toposort(circuit: CircuitRc) -> Schedule {
     let toposorted = toposort_circuit(circuit.clone());
-    order_to_schedule(
+    let mut result = order_to_schedule(
         &toposorted,
         &vec![],
         &vec![],
         HashSet::from([circuit.info().hash]),
         get_compatible_dtype(&circuit),
-    )
+    );
+    result.output_circuit = Some(circuit);
+    result
 }
 
 #[pyfunction]
@@ -659,5 +814,5 @@ pub fn scheduled_evaluate(
     settings: OptimizationSettings,
 ) -> Result<Tensor, SchedulingError> {
     let schedule = circuit_to_schedule(circuit.clone(), settings)?;
-    Ok(evaluate_schedule(schedule)[&circuit.info().hash].clone())
+    Ok(evaluate_schedule(&schedule)[&circuit.info().hash].clone())
 }
