@@ -119,6 +119,9 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
 
     m.add_class::<circuit::PyGFSpecShapeGetter>()?;
 
+    m.add_class::<circuit::scheduled_execution::Schedule>()?;
+    m.add_class::<circuit::scheduled_execution::ScheduleStats>()?;
+
     m.add_function(wrap_pyfunction!(add_collapse_scalar_inputs, m)?)?;
     m.add_function(wrap_pyfunction!(add_deduplicate, m)?)?;
     m.add_function(wrap_pyfunction!(remove_add_few_input, m)?)?;
@@ -185,6 +188,14 @@ fn rust_circuit(py: Python<'_>, m: &PyModule) -> PyResult<()> {
     m.add_function(wrap_pyfunction!(scheduled_evaluate, m)?)?;
     m.add_function(wrap_pyfunction!(optimize_and_evaluate, m)?)?;
     m.add_function(wrap_pyfunction!(optimize_and_evaluate_many, m)?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit::circuit_optimizer::optimize_to_schedule,
+        m
+    )?)?;
+    m.add_function(wrap_pyfunction!(
+        circuit::circuit_optimizer::optimize_to_schedule_many,
+        m
+    )?)?;
     m.add_function(wrap_pyfunction!(flat_concat, m)?)?;
     m.add_function(wrap_pyfunction!(deep_heuristic_nest_adds, m)?)?;
     m.add_function(wrap_pyfunction!(concat_fuse, m)?)?;
