@@ -18,8 +18,8 @@ lazy_static! {
         PyModule::from_code(
             py,
             include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/rust_circuit_type_utils.py")),
-            "rust_circuit_type_utils.py",
-            "rust_circuit_type_utils",
+            concat!(env!("CARGO_MANIFEST_DIR"), "/rust_circuit_type_utils.py"),
+            "rust_circuit.rust_circuit_type_utils",
         )
         .unwrap() // TODO: fail more gracefully?
         .into()
@@ -55,15 +55,17 @@ lazy_static! {
         Python::with_gil(|py| PYTHON_UTILS.getattr(py, "einsum").unwrap().extract(py).unwrap());
     pub static ref PY_MAKE_DIAGONAL: PyObject=
         Python::with_gil(|py| PYTHON_UTILS.getattr(py, "make_diagonal").unwrap().extract(py).unwrap());
+    pub static ref PY_MAKE_BYTES: PyObject=
+        Python::with_gil(|py| PYTHON_UTILS.getattr(py, "construct_bytes").unwrap().extract(py).unwrap());
 }
 
 lazy_static! {
     static ref PYCIRCUIT_MODULE: Py<PyModule> = Python::with_gil(|py| {
         PyModule::from_code(
             py,
-            "from interp.circuit import computational_node, constant; import interp.circuit.interop_rust.interop_rust as interop; import einops; import interp.circuit.circuit_compiler.util as circ_compiler_util",
-            "py_circuit_interop.py",
-            "py_circuit_interop",
+            include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/py_circuit_interop.py")),
+            concat!(env!("CARGO_MANIFEST_DIR"), "/py_circuit_interop.py"),
+            "rust_circuit.py_circuit_interop",
         )
         .unwrap() // TODO: fail more gracefully?
         .into()
@@ -88,8 +90,8 @@ lazy_static! {
         PyModule::from_code(
             py,
             include_str!(concat!(env!("CARGO_MANIFEST_DIR"), "/tensor_hash.py")),
-            "tensor_hash.py",
-            "tensor_hash",
+            concat!(env!("CARGO_MANIFEST_DIR"), "/tensor_hash.py"),
+            "rust_circuit.tensor_hash",
         )
         .unwrap() // TODO: fail more gracefully?
         .into()
