@@ -87,25 +87,26 @@ pub struct Slice {
 impl Slice {
     pub fn to_unsigned_loc(x: i64, l: usize) -> usize {
         if x < 0 {
-            l.saturating_sub((-x) as usize)
+            l.checked_sub((-x) as usize).unwrap()
         } else {
-            l.min(x as usize)
+            assert!(x <= l as i64);
+            x as usize
         }
     }
 
-    fn start_u(self, l: usize) -> usize {
+    pub fn start_u(self, l: usize) -> usize {
         self.start.map(|x| Self::to_unsigned_loc(x, l)).unwrap_or(0)
     }
 
-    fn stop_u(self, l: usize) -> usize {
+    pub fn stop_u(self, l: usize) -> usize {
         self.stop.map(|x| Self::to_unsigned_loc(x, l)).unwrap_or(l)
     }
 
-    fn size(self, l: usize) -> usize {
+    pub fn size(self, l: usize) -> usize {
         self.stop_u(l).saturating_sub(self.start_u(l))
     }
 
-    fn canonicalize(self, l: usize) -> Self {
+    pub fn canonicalize(self, l: usize) -> Self {
         Self {
             start: Some(self.start_u(l) as i64),
             stop: Some(self.stop_u(l) as i64),
@@ -127,7 +128,7 @@ impl Slice {
         }
     }
 
-    fn is_identity(self, l: usize) -> bool {
+    pub fn is_identity(self, l: usize) -> bool {
         self.start_u(l) == 0 && self.stop_u(l) == l
     }
 }
@@ -423,7 +424,12 @@ impl TensorIndex {
             zip(&self.0, shape)
                 .map(|(idx, &l)| match idx {
                     TensorAxisIndex::Single(i) => {
-                        assert!(*i < l as i64 && *i >= -(l as i64));
+                        assert!(
+                            *i < l as i64 && *i >= -(l as i64),
+                            "single oob idx {} length {}",
+                            i,
+                            l
+                        );
                         TensorAxisIndex::Single((*i + l as i64) % l as i64)
                     }
                     TensorAxisIndex::Tensor(t) => TensorAxisIndex::Tensor(t.clone()), // not bothering to canon tensor for now
