@@ -204,20 +204,17 @@ impl<T> AsOp<T> for T {
 
 pub fn mapping_until_end<T: Eq + Hash + Clone>(x: &T, mapping: &HashMap<T, T>) -> T {
     let mut result = x.clone();
-    for i in 0..1000 {
+    for _ in 0..1000 {
         match mapping.get(&result) {
             None => {
-                break;
+                return result;
             }
             Some(next) => {
                 result = next.clone();
             }
         }
-        if i == 1000 - 1 {
-            panic!("mapping_until_end didnt finish");
-        }
     }
-    result
+    panic!("mapping_until_end didnt finish");
 }
 
 pub fn apply_fn_until_same<T: Eq + Hash + Clone, F>(x: &T, f: F) -> T
